from fileinput import filename
import fnmatch
from typing import Optional
from dataclasses import dataclass
from subprocess import run, PIPE
from .exceptions import LoggedError


@dataclass
class Commit:
    sha: str
    first_parent: Optional[str] = None
    second_parent: Optional[str] = None

    @property
    def is_empty_merge(self) -> bool:
        """Check if commit is a merge that could have been fast-forward."""
        if self.first_parent and self.second_parent:
            # If the first parent is an ancestor of the second, then the head branch
            # was not behind the base branch by any commits, meaning a fast-forward was possible
            return _is_ancestor(self.first_parent, self.second_parent)

        # Not a merge
        return False


class DirtyWorkingTreeError(LoggedError):
    def __init__(self) -> None:
        super().__init__("Working tree is dirty.")


class NoRepoError(LoggedError):
    def __init__(self) -> None:
        super().__init__("Not a git repository or no remote 'origin' set.")


def ensure_repo() -> None:
    cmd = ["git", "remote", "get-url", "origin"]
    result = run(cmd, stdout=PIPE, stderr=PIPE)
    if result.returncode != 0:
        raise NoRepoError()


def ensure_clean_working_tree() -> None:
    cmd = ["git", "status", "--short"]
    result = run(cmd, stdout=PIPE, stderr=PIPE)
    output = result.stdout.decode().strip()
    if len(output) > 0:
        raise DirtyWorkingTreeError()


def unshallow() -> None:
    cmd = ["git", "fetch", "--unshallow"]
    run(cmd, check=True)


def get_head() -> Commit:
    """Get the head commit with its parent(s)"""
    cmd = ["git", "log", "--pretty=%h,%p", "-n", "1"]
    result = run(cmd, stdout=PIPE)
    print("git log stdout:", result.stdout)
    head, parents = result.stdout.decode().strip().split(",")
    return Commit(head, *parents.split(" "))


def get_repo_name() -> str:
    cmd = ["git", "remote", "get-url", "origin"]
    result = run(cmd, stdout=PIPE)
    url = result.stdout.decode().strip()
    name = url.split("/")[-1].split(".")[0].replace("github_aaa-ncnu_", "")
    return name


def changed_since(start_commit: str, file_pattern: str) -> bool:
    cmd = ["git", "diff", "--name-only", start_commit, "HEAD"]
    result = run(cmd, stdout=PIPE)
    filenames = result.stdout.decode().strip().split("\n")
    matches = fnmatch.filter(filenames, file_pattern)
    return len(matches) > 0


def _is_ancestor(commit_a: str, commit_b: str) -> bool:
    """Checks if commit_a is an ancestor of commit_b"""
    cmd = ["git", "merge-base", "--is-ancestor", commit_a, commit_b]
    result = run(cmd, stdout=PIPE)
    print("is ancestor:", commit_a, commit_b, result.returncode)
    return result.returncode == 0
